/* poppasswd.c
   -----------
   Update passwords on the UNIX system with the protocol introduced by the
   Eudora program and now used by most POP clients. Keeps a log of change
   attempts in syslog, but suppresses recording of passwords given by the
   client. Must be owned by root but should execute NON-PRIVILEGED and
   CHROOTED.
   It can be started with an entry in /etc/inetd.conf such as the following:
 
   poppass stream tcp nowait root S4MAILHOME/pop_delivery/bin/poppasswd
                 poppasswd S4POPADMINNAME:S4POPGNAME:S4MAILHOME/pop_delivery

           and in /etc/services:

             poppass        106/tcp

   Note: I believe that the original of this program was written by
         Daniel L. Leavitt (dll@mitre.org). At least this is what file
         pwserve-1 on ftp.qualcomm.com refered to at the time I picked this
         up. I have modified the code heavily for integration into S4.

         I THINK the original code did not have any particular licensing
         rights. If I am wrong, please let me know.

   Last known maintainer:          Marc Chatel   (mchatel@dial.oleane.com)
                                   September 1996
*/

#include <stdio.h>
#include <string.h>
#include <pwd.h>
#include <signal.h>
#include <unistd.h>
#include <varargs.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/file.h>
#include <time.h>
#include <syslog.h>
#include "ipsyslog.h"

#define PF_PLATE        "/tmp/ptmpXXXXX" /* original temporary file */
#define PF_TEMP         "/tmp/ptmp"      /* lock file */
#define PF_DIR          "/poparea"       /* Directory for the password file */
#define PF_NAME         "poppasswd"      /* Name of the password file */

#define PW_CHANGED      0               /* a OK */
#define PW_CANTLOCK     -1              /* error code -- locking failed */
#define PW_CANTOPEN     -2              /* error code -- open failed */
#define PW_CANTRENAME   -3              /* error code -- rename failed */
#define PW_NOCHANGE     -4              /* error code -- no passwd entry */

#define SUCCESS 1
#define FAILURE 0
#define BUFSIZE 255
#define MAXARGS 32
#define SUPPRESS 1

#define FI_NULL ((FILE *) NULL)                /* NULL file pointer */
#define TI_NULL ((time_t *) NULL)      /* NULL time pointer */


/*  library functions */

extern char *mktemp();                  /* makes a temp file name */

/* global variables */

char *pf_name = PF_NAME;               /* password file name */
char *pf_tnam;                          /* name of temporary file */
FILE *pf_fp;                            /* pointer to real password file */
FILE *pf_tfp;                           /* pointer to temporary file */
FILE *log;
time_t gNow;
char gDate[BUFSIZE];


void UpdateTime()
{
   struct tm *date;
       
   gNow = time( NULL );            /* get current calendar time */
   date = localtime( &gNow );      /* convert it to a broken-down time */
   strftime( gDate, BUFSIZE, "%b %d %H:%M:%S", date );
}


void WhoAreYou()
{
   struct sockaddr_in   bobby;
   int                  bobbyLen = sizeof(struct sockaddr);
   struct passwd       *pw;
   char                *ttyname();

   if (isatty(fileno(stdin)))
   {
      pw = getpwuid(getuid());
      ip_syslog( LOG_INFO, "Connection on %s by %s", 
                           ttyname(fileno(stdin)), pw->pw_name);
      return;
   }

   if (lseek(fileno(stdin), 0, L_INCR) >= 0)
   {
      pw = getpwuid(getuid());
      ip_syslog( LOG_INFO, "Connection on FILE by %s", pw->pw_name);
      return;
   }

   if (getpeername(fileno(stdin), (struct sockaddr *)&bobby, &bobbyLen) < 0)
   {
      pw = getpwuid(getuid());
      ip_syslog( LOG_INFO, "Connection on PIPE by %s", pw->pw_name);
      return;
   }

#if defined(sparc) && __GNUC__ == 1 
   ip_syslog( LOG_INFO, "NET connection from  %s", inet_ntoa(&bobby.sin_addr));
#else
   ip_syslog( LOG_INFO, "NET connection from %s", inet_ntoa( bobby.sin_addr ));
#endif
}


void WriteToClient( our_msg )
char *our_msg;
{
   printf( "%s\r\n", our_msg );
/* ip_syslog( LOG_DEBUG, "Server> %s", our_msg ); */
   fflush(stdout);
}


void ReadFromClient( line , suppress )
char *line;
int suppress;
{
   char *sp;
   int   i;

   strcpy(line, "" );
   fgets( line, BUFSIZE, stdin );
   if (( sp = strchr(line, '\n')) != NULL ) *sp = '\0'; 
   if (( sp = strchr(line, '\r')) != NULL ) *sp = '\0';
/* if (suppress) ip_syslog( LOG_DEBUG, "Client> ..." );
   else          ip_syslog( LOG_DEBUG, "Client> %s", line ); */
}


int chkPass( user, pass, pw )
char *user;
char *pass;
struct passwd *pw;
{
   /*  Compare the supplied password with the password file entry */

   if (strcmp(crypt(pass, pw->pw_passwd), pw->pw_passwd) != 0)
        return ( FAILURE );
   else return ( SUCCESS );
}


void makesalt( c )
char c[2];                      /* salt characters */
{
   register long salt;             /* used to compute a salt */
   register int i;                 /* counter in a for loop */

   /* just mix a few things up for the salt ... no rhyme or reason */

   salt = (((long) time(TI_NULL))&0x3f) | (getpid() << 5);

   /* use the bottom 12 bits and map them into the legal alphabet */

   for (i = 0; i < 2; i++)
   {
      c[i] = (salt & 0x3f) + '.';
      if (c[i] > '9') c[i] += 7;
      if (c[i] > 'Z') c[i] += 6;
      salt >>= 6;
   }
}


/* signal handling routines */

void (*trap[NSIG])();           /* values returned by signal() */


/* disable ALL signal trapping */

sigoff()
{
        register int i;         /* counter in a for loop */

        for(i = 0; i < NSIG; i++) (void) signal(i, SIG_IGN);
}


/* restore to the initial setting all signals */

sigon()
{
        register int i;         /* counter in a for loop */

        for(i = 0; i < NSIG; i++) (void) signal(i, trap[i]);
}


int UpdatePasswordFile(p)
struct passwd *p;               /* new password structure */
{
   char           nbuf[BUFSIZ];          /* buffer for passwords being read */
   char           obuf[BUFSIZ];          /* buffer for new entry */
   register char *q, *s;                 /* used for quick name comparison */
   int            retval = PW_NOCHANGE;  /* return value */

   /* disable ALL signals at this point */
   sigoff();

   /* open the temporary password file */
   umask(022);
   pf_tnam = mktemp(PF_PLATE);

   if ((pf_tfp = fopen(pf_tnam, "w")) == FI_NULL)
   {
      retval = PW_CANTOPEN;
      goto cantlock;
   }

   /* lock the password file */

   if (link(pf_tnam, PF_TEMP) < 0)
   {
      retval = PW_CANTLOCK;
      goto cantlock;
   }
        
   /* copy the new password structure */

   sprintf(obuf, "%s:%s:%d:%d:%s:%s:%s\n",
                 p->pw_name, p->pw_passwd, p->pw_uid, p->pw_gid,
                 p->pw_gecos, p->pw_dir, p->pw_shell );

   /* open the password file */

   if ((pf_fp = fopen(pf_name, "r")) == FI_NULL)
   {
      retval = PW_CANTOPEN;
      goto getout;
   }

   /* copy the password file into the temporary one */

   while (fgets(nbuf, BUFSIZ, pf_fp) != NULL)
   {
      for (s = nbuf, q = p->pw_name; *s && *s != ':'; s++, q++)
         if (*s != *q) break;

      if (*s == ':' && *q == '\0')
      {
         fputs(obuf, pf_tfp);
         retval = PW_CHANGED;
      }
      else fputs(nbuf, pf_tfp);
   }

   if (retval == PW_NOCHANGE) goto getout;

   /* close the temporary file and the real one */

   (void) fclose(pf_tfp);
   pf_tfp = FI_NULL;
   (void) fclose(pf_fp);
   pf_fp = FI_NULL;

   /* now relink; note the lock file is still there */

   if ( unlink(pf_name) >= 0 && link(pf_tnam, pf_name) >= 0
                             && unlink(pf_tnam) >= 0)
   {
        retval = PW_CHANGED;
   }
   else retval = PW_CANTRENAME;

getout:
   /* Only remove lock file if this program obtained it */

   (void) unlink(PF_TEMP); 

cantlock:
   /* some clean up */

   if (pf_tfp != FI_NULL) (void) fclose(pf_tfp);
   if (pf_fp != FI_NULL)  (void) fclose(pf_fp);
   (void) unlink(pf_tnam);

   sigon(); /* re-enable ALL signals at this point */

   return(retval);
}


int setPass( user, pass, newpass, pw )
char *user;
char *pass;
char *newpass;
struct passwd *pw;
{
   char saltc[2];          /* the password's salt */

   makesalt(saltc);
   strcpy( pw->pw_passwd, crypt(newpass, saltc));

   if ( UpdatePasswordFile( pw ) == PW_CHANGED ) return (SUCCESS);
   else                                          return (FAILURE);
}

#define NUM_VALID_CMDS   3
char *cmdlist[NUM_VALID_CMDS] = { "USER", "PASS", "NEWPASS" };
#define USER_CMD     0
#define PASS_CMD     1
#define NEWPASS_CMD  2

int calc_cmd( line_in, first_parm_ptr )
char *line_in;
char **first_parm_ptr;
{
   int our_rc = -1;
   char *tmp;

   *first_parm_ptr = NULL;

   if (line_in == NULL)     return our_rc;

   while (((*line_in) == ' ') || ((*line_in) == '\t')) line_in++;

   if ((*line_in) == '\0')  return our_rc;

   tmp = line_in;
   while ((*tmp != ' ') && (*tmp != '\t') && (*tmp != '\0')) tmp++;

   if (*tmp != '\0')    *(tmp++) = '\0';

   /* Now we compare the string pointed to by line_in to our
      array of commands */

   {
      int tmp_idx = 0;

      for (tmp_idx = 0; tmp_idx < NUM_VALID_CMDS; tmp_idx++)
      {
         if (strcasecmp( cmdlist[tmp_idx], line_in ) == 0)
         {
/*          ip_syslog( LOG_DEBUG, "Found command number %d", tmp_idx ); */
            our_rc = tmp_idx; break;
         }
      }
   }

   if (*tmp == '\0')   return our_rc;
   while ((*tmp == ' ') || (*tmp == '\t')) tmp++;

   if (*tmp != '\0')  *first_parm_ptr = tmp;
   while ((*tmp != ' ') && (*tmp != '\t') && (*tmp != '\0') &&
         (*tmp != '\r') && (*tmp != '\n')) tmp++;
   *tmp == '\0';
   return our_rc;
}

int main(argc, argv)
int argc;
char *argv[];
{
   char line[BUFSIZE];
   char user[BUFSIZE];
   char pass[BUFSIZE];
   char newpass[BUFSIZE];
   struct passwd *pw = NULL;
   char *tmp_parm;

   line[0] = '\0'; user[0] = '\0'; pass[0] = '\0'; newpass[0] = '\0';
   startup_ip_syslog( "poppasswd", LOG_MAIL );

   if (argc != 2)
   {
      syslog( LOG_ERR,
              "Need a drop_priv parameter uname:gname:rootdir" );
      exit(1);
   }

   if (!drop_priv_option( argv[1] ))
   {
      syslog( LOG_ERR,
              "drop_priv failed!" );
      exit(1);
   }

   chdir( PF_DIR );

#if 0
   The server's responses should be like an FTP server's responses; 
   1xx for in progress, 2xx for success, 3xx for more information
   needed, 4xx for temporary failure, and 5xx for permanent failure.  
   Putting it all together, here's a sample conversation:

   S: 200 hello\r\n
   E: user yourloginname\r\n
   S: 300 please send your password now\r\n
   E: pass yourcurrentpassword\r\n
   S: 200 My, that was tasty\r\n
   E: newpass yournewpassword\r\n
   S: 200 Happy to oblige\r\n
   E: quit\r\n
   S: 200 Bye-bye\r\n
   S: <closes connection>
   E: <closes connection>
#endif

   WhoAreYou();
   WriteToClient( "200 hello, who are you?" );
   ReadFromClient( line, 0 );
   
   if (calc_cmd( line, &tmp_parm ) != USER_CMD)
   {
      WriteToClient( "500 must be USER cmd" );
      ip_syslog( LOG_ERR, "ERR Expected USER command" );
      exit(1);
   }
   
   if ((tmp_parm == NULL) || (strlen(tmp_parm) == 0))
   {
      WriteToClient( "500 username required." );
      ip_syslog( LOG_ERR, "ERR USER command without parameter" );
      exit(1);
   }

   strcpy( user, tmp_parm );

   WriteToClient( "200 your password please." );
   ReadFromClient( line, SUPPRESS );

   if (calc_cmd( line, &tmp_parm ) != PASS_CMD)
   {
      WriteToClient( "500 must be PASS cmd" );
      ip_syslog( LOG_ERR, "ERR user %s, expected PASS command", user );
      exit(1);
   }

   if ((tmp_parm == NULL) || (strlen(tmp_parm) == 0))
   {
      WriteToClient( "500 password required." );
      ip_syslog( LOG_ERR,
             "ERR user %s, PASS command without parameter", user );
      exit(1);
   }

   strcpy( pass, tmp_parm );

   WriteToClient( "200 your new password please." );
   ReadFromClient( line, SUPPRESS );

   if (calc_cmd( line, &tmp_parm ) != NEWPASS_CMD)
   {
      WriteToClient( "500 must be NEWPASS cmd" );
      ip_syslog( LOG_ERR, "ERR user %s, Expected NEWPASS command",
                          user );
      exit(1);
   }

   /* new pass required */
   if ((tmp_parm == NULL) || (strlen(tmp_parm) == 0))
   {
      WriteToClient( "500 new password required." );
      ip_syslog( LOG_ERR,
                 "ERR user %s, NEWPASS command without parameter",
                 user );
      exit(1);
   }

   strcpy( newpass, tmp_parm );

   /* new pass must be 6 char or longer */
   if (strlen(newpass) < 6)
   {
      WriteToClient( "500 New password too short" );
      ip_syslog( LOG_ERR, "ERR user %s, New password too short",
                          user );
      exit(1);
   }

   /* new pass must be different from old pass */
   if (!strcmp(pass, newpass))
   {
      WriteToClient( "500 New password must be different." );
      ip_syslog( LOG_ERR, "ERR user %s, Tried to keep same password",
                          user );
      exit(1);
   }

   {
      char tmpbuf[500];

      strcpy( tmpbuf, PF_DIR );
      strcat( tmpbuf, "/" );
      strcat( tmpbuf, PF_NAME );
      setpwfile( tmpbuf );
   }

   /* test for valid user */

   if ((pw=getpwnam(user)) == NULL)
   {
      WriteToClient( "500 Password change failed");
      ip_syslog( LOG_ERR, "ERR user %s unknown", user );
      exit(1);
   }

   /* authenticate user */
   if (chkPass( user, pass, pw ) == FAILURE)
   {
      WriteToClient( "500 Password change failed");
      ip_syslog( LOG_ERR, "ERR user %s supplied bad password", user );
      exit(1);
   }
                
   if (setPass( user, pass, newpass, pw ) == FAILURE)
   { 
      WriteToClient( "500 Password change failed" );
      ip_syslog( LOG_ERR, "ERR user %s failed to change password",
                          user );
      exit(1);
   }

   WriteToClient( "200 Password changed, thank-you." );
   ip_syslog( LOG_INFO, "INFO user %s changed password OK", user );
   return 0;
}
